
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_dynamic_vlsn.c                                      */
/*  DESCRIPTION:  Dynamic resource VLSN solving                              */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#include <limits.h>
#define is_digit(x) ((x) >= '0' && (x) <= '9')
#define bool_show(x) ((x) ? "true" : "false")
#define max(x, y) ((x) > (y) ? (x) : (y))
#define min(x, y) ((x) < (y) ? (x) : (y))

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0
#define DEBUG4 0
#define DEBUG5 0	/* cluster */


/*****************************************************************************/
/*                                                                           */
/* Submodule "type declarations"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  Type KHE_RESOURCE_SELECT - a sequence of resource sets                   */
/*                                                                           */
/*****************************************************************************/
typedef HA_ARRAY(KHE_RESOURCE) ARRAY_KHE_RESOURCE;
typedef HA_ARRAY(KHE_RESOURCE_SET) ARRAY_KHE_RESOURCE_SET;

typedef enum {
  KHE_RESOURCE_SELECT_ENUMERATED,
  KHE_RESOURCE_SELECT_CHOOSE,
  KHE_RESOURCE_SELECT_CLUSTER,
} KHE_RESOURCE_SELECT_KIND;

typedef enum {
  KHE_UNDERLOADED	= 0,
  KHE_MAXIMAL		= 1,
  KHE_OVERLOADED	= 2
} KHE_RESOURCE_LOAD;

typedef struct khe_resource_select_rec {
  KHE_RESOURCE_TYPE		resource_type;
  KHE_RESOURCE_SELECT_KIND	kind;
  ARRAY_KHE_RESOURCE_SET	free_resource_sets;
  union {
    struct {
      ARRAY_KHE_RESOURCE_SET	resource_sets;
    } en;
    struct {
      int			choose_len;		/* if choose(x)      */
      ARRAY_KHE_RESOURCE	randomized_resources;	/* all resources     */
      KHE_RESOURCE_SET		scratch_resource_set;
    } ch;
    struct {
      char			*constraint_str;
      /* int			resource_count; */
      ARRAY_KHE_RESOURCE	load_class[KHE_OVERLOADED + 1];
      ARRAY_KHE_RESOURCE_SET	resource_sets;
    } cl;
  } u;
} *KHE_RESOURCE_SELECT;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_INTERVAL_SET - a set of intervals                               */
/*                                                                           */
/*****************************************************************************/
typedef HA_ARRAY(KHE_INTERVAL) ARRAY_KHE_INTERVAL;

typedef struct khe_interval_set_rec {
  ARRAY_KHE_INTERVAL	intervals;
} *KHE_INTERVAL_SET;

typedef HA_ARRAY(KHE_INTERVAL_SET) ARRAY_KHE_INTERVAL_SET;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DAY_SELECT - a day select, usually a sequence of interval sets  */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_day_select_rec {
  KHE_FRAME			days_frame;
  int				choose_len;		/* if choose(x)      */
  ARRAY_KHE_INTERVAL_SET	interval_sets;
  KHE_INTERVAL_SET		single_interval_set;
} *KHE_DAY_SELECT;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_SOLVE_ARGUMENTS - arguments to a solve                          */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_solve_arguments_rec {
  KHE_RESOURCE_SELECT	resources;			/* R                 */
  KHE_DAY_SELECT	days;				/* D                 */
  bool			priqueue;			/* Q                 */
  bool			extra_selection;		/* E                 */
  bool			expand_by_shifts;		/* S                 */
  bool			shift_pairs;			/* P                 */
  bool			correlated_exprs;		/* C                 */
  int			daily_expand_limit;		/* L                 */
  int			daily_prune_trigger;		/* T                 */
  int			resource_expand_limit;		/* M                 */
  int			dom_approx;			/* A                 */
  KHE_DRS_DOM_KIND	main_dom_kind;			/* K                 */
  bool			cache;				/* J                 */
  KHE_DRS_DOM_KIND	cache_dom_kind;			/* J                 */
} *KHE_SOLVE_ARGUMENTS;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_OPTIMAL_RESULT - one optimal result                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_optimal_result_rec {
  KHE_RESOURCE_SET	resource_set;
  KHE_INTERVAL_SET	day_set;
  KHE_COST		cost;
} *KHE_OPTIMAL_RESULT;

typedef HA_ARRAY(KHE_OPTIMAL_RESULT) ARRAY_KHE_OPTIMAL_RESULT;


/*****************************************************************************/
/*                                                                           */
/*  KHE_OPTIMAL_RESULT_SET - a set of optimal results                        */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_optimal_result_set_rec {
  ARRAY_KHE_OPTIMAL_RESULT	results;
} *KHE_OPTIMAL_RESULT_SET;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DIM_POS_OPT - one option of one position of one dimension       */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_drs_dim_pos_opt_rec {
  char				tag;		/* capital letter */
  union {
    bool			bool_val;
    int				int_val;
    KHE_DRS_DOM_KIND		dom_kind_val;
    KHE_RESOURCE_SELECT		resource_select_val;
    KHE_DAY_SELECT		day_select_val;
  } u;
} *KHE_DIM_POS_OPT;

typedef HA_ARRAY(KHE_DIM_POS_OPT) ARRAY_KHE_DIM_POS_OPT;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DIM_POS - one position of one dimension                         */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_dim_pos_rec {
  ARRAY_KHE_DIM_POS_OPT		options;
} *KHE_DIM_POS;

typedef HA_ARRAY(KHE_DIM_POS) ARRAY_KHE_DIM_POS;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_DIM - one dimension                                             */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_dim_rec {
  bool				days_dim;		/* true if dim is Z  */
  ARRAY_KHE_DIM_POS		dim_poses;
} *KHE_DIM;

typedef HA_ARRAY(KHE_DIM) ARRAY_KHE_DIM;


/*****************************************************************************/
/*                                                                           */
/*  Type KHE_RUNNER - controller for solving and testing                     */
/*                                                                           */
/*****************************************************************************/

typedef struct khe_runner_rec {
  HA_ARENA			arena;
  KHE_SOLN			soln;
  KHE_RESOURCE_TYPE		resource_type;
  KHE_OPTIONS			options;
  KHE_FRAME			days_frame;
  KHE_RANDOM_GENERATOR		rgen;
  KHE_SOLVE_ARGUMENTS		solve_args;
  KHE_DYNAMIC_RESOURCE_SOLVER	solver;
  ARRAY_KHE_DIM			test_dims;
  KHE_OPTIMAL_RESULT_SET	optimal_results;
} *KHE_RUNNER;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "basic parsing"                                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheParseHasEnded(char **str)                                        */
/*                                                                           */
/*  Return true if the parse of *str has ended.                              */
/*                                                                           */
/*****************************************************************************/

static bool KheParseHasEnded(char **str)
{
  return **str == '\0';
}


/*****************************************************************************/
/*                                                                           */
/*  char KheCharParse(char **str)                                            */
/*                                                                           */
/*  Read the next character from *str, return it, and skip over it.          */
/*                                                                           */
/*****************************************************************************/

static char KheCharParse(char **str)
{
  char res;
  res = **str;
  (*str)++;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheStrSkip(char **str, char *val)                                   */
/*                                                                           */
/*  Skip val and return true if successful, else change nothing and          */
/*  return false.                                                            */
/*                                                                           */
/*****************************************************************************/

static bool KheStrSkip(char **str, char *val)
{
  char *p, *q;
  for( p = *str, q = val;  *q != '\0';  p++, q++ )
    if( *p != *q )
      return false;
  *str = p;
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheBoolParse(char **str)                                            */
/*                                                                           */
/*  Read the next bool from *str, return it, and skip over it.               */
/*                                                                           */
/*****************************************************************************/

static bool KheBoolParse(char **str)
{
  if( KheStrSkip(str, "true") )
    return true;
  else if( KheStrSkip(str, "false") )
    return false;
  else
  {
    HnAbort("error where true or false expected in rs_drs_test option");
    return false;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  int KheIntParse(char **str)                                              */
/*                                                                           */
/*  Read the next int from *str, return it, and skip over it.                */
/*                                                                           */
/*****************************************************************************/

static int KheIntParse(char **str)
{
  int res;
  if( sscanf(*str, "%d", &res) != 1 )
    HnAbort("error where integer expected in rs_drs_test option");
  while( is_digit(**str) )
    (*str)++;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheStringParse(char **str, char stop)                              */
/*                                                                           */
/*  Read the next string from *str, return it, and skip over it.             */
/*                                                                           */
/*****************************************************************************/

static char *KheStringParse(char **str, char stop)
{
  static char buff[200];  char *p, *q;
  for( p = *str, q = buff; *p != '\0' && *p != stop && q < &buff[199]; p++, q++)
    *q = *p;
  *q = '\0';
  *str = p;
  return buff;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_RESOURCE_SET" - additional to khe_platform.h              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SET KheResourceSetParse(char **str, KHE_RESOURCE_TYPE rt,   */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Parse a resource set (represented by a comma-separated sequence of       */
/*  integer indexes into rt).                                                */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SET KheResourceSetParse(char **str, KHE_RESOURCE_TYPE rt,
  HA_ARENA a)
{
  KHE_RESOURCE_SET res;  KHE_RESOURCE r;  int val;
  res = KheResourceSetMake(rt, a);
  do
  {
    val = KheIntParse(str);
    if( 0 <= val && val < KheResourceTypeResourceCount(rt) )
    {
      r = KheResourceTypeResource(rt, val);
      KheResourceSetAddResource(res, r);
    }
  } while( KheStrSkip(str, ",") );
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceSetAddDisplay(KHE_RESOURCE_SET rset, HA_ARRAY_NCHAR *res)*/
/*                                                                           */
/*  Add a display of rset to *res.                                           */
/*                                                                           */
/*****************************************************************************/

static void KheResourceSetAddDisplay(KHE_RESOURCE_SET rset, HA_ARRAY_NCHAR *res)
{
  int i;  KHE_RESOURCE r;
  for( i = 0;  i < KheResourceSetResourceCount(rset);  i++ )
  {
    if( i > 0 )
      HnStringAdd(res, ",");
    r = KheResourceSetResource(rset, i);
    HnStringAdd(res, "%s", KheResourceId(r));
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_RESOURCE_SELECT"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SELECT KheResourceSelectMakeEnumerated(KHE_RESOURCE_TYPE rt,*/
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Make an enumerated resource select object with these attributes.         */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SELECT KheResourceSelectMakeEnumerated(KHE_RESOURCE_TYPE rt,
  HA_ARENA a)
{
  KHE_RESOURCE_SELECT res;
  HaMake(res, a);
  res->resource_type = rt;
  res->kind = KHE_RESOURCE_SELECT_ENUMERATED;
  HaArrayInit(res->free_resource_sets, a);
  HaArrayInit(res->u.en.resource_sets, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SELECT KheResourceSelectMakeChoose(KHE_RESOURCE_TYPE rt,    */
/*    int choose_len, HA_ARENA a)                                            */
/*                                                                           */
/*  Make a choose resource select object with these attributes.              */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SELECT KheResourceSelectMakeChoose(KHE_RESOURCE_TYPE rt,
  int choose_len, HA_ARENA a)
{
  KHE_RESOURCE_SELECT res;  KHE_RESOURCE r;  int i;

  /* reduce choose_len if too large */
  HnAssert(choose_len >= 1, "KheResourceSelectMake internal error 1");
  if( choose_len > KheResourceTypeResourceCount(rt) )
    choose_len = KheResourceTypeResourceCount(rt);

  /* make the object */
  HaMake(res, a);
  res->resource_type = rt;
  res->kind = KHE_RESOURCE_SELECT_CHOOSE;
  HaArrayInit(res->free_resource_sets, a);
  res->u.ch.choose_len = choose_len;
  HaArrayInit(res->u.ch.randomized_resources, a);
  for( i = 0;  i < KheResourceTypeResourceCount(rt);  i++ )
  {
    r = KheResourceTypeResource(rt, i);
    HaArrayAddLast(res->u.ch.randomized_resources, r);
  }
  res->u.ch.scratch_resource_set = KheResourceSetMake(rt, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SELECT KheResourceSelectMakeCluster(KHE_RESOURCE_TYPE rt,   */
/*    char *constraint_str, int resource_count, HA_ARENA a)                  */
/*                                                                           */
/*  Make a cluster resource select object with these attributes.             */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SELECT KheResourceSelectMakeCluster(KHE_RESOURCE_TYPE rt,
  char *constraint_str, /* int resource_count, */ HA_ARENA a)
{
  KHE_RESOURCE_SELECT res;  int i;
  HaMake(res, a);
  res->resource_type = rt;
  res->kind = KHE_RESOURCE_SELECT_CLUSTER;
  HaArrayInit(res->free_resource_sets, a);
  res->u.cl.constraint_str = constraint_str;
  /* res->u.cl.resource_count = resource_count; */
  for( i = KHE_UNDERLOADED;  i <= KHE_OVERLOADED;  i++ )
    HaArrayInit(res->u.cl.load_class[i], a);
  HaArrayInit(res->u.cl.resource_sets, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int FindIndex(KHE_RESOURCE_SELECT rs, KHE_RESOURCE over_r,               */
/*    KHE_RESOURCE_LOAD r_load)                                              */
/*                                                                           */
/*  Find an index into rs->load_class[r_load] which points to a resource     */
/*  whose index is close to over_r's index.                                  */
/*                                                                           */
/*  Here rs->load_class[r_load] must contain at least one element.           */
/*                                                                           */
/*****************************************************************************/

static int FindIndex(KHE_RESOURCE_SELECT rs, KHE_RESOURCE over_r,
  KHE_RESOURCE_LOAD r_load)
{
  int over_r_index, i;  KHE_RESOURCE r;
  HnAssert(HaArrayCount(rs->u.cl.load_class[r_load]) > 0,
    "FindIndex internal error");
  over_r_index = KheResourceResourceTypeIndex(over_r);
  HaArrayForEachReverse(rs->u.cl.load_class[r_load], r, i)
    if( KheResourceResourceTypeIndex(r) < over_r_index )
      return i;
  return 0;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SET KheResourceSetGet(KHE_RESOURCE_SELECT rs, HA_ARENA a)   */
/*                                                                           */
/*  Get a resource set from rs's free list or from a.                        */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SET KheResourceSetGet(KHE_RESOURCE_SELECT rs, HA_ARENA a)
{
  KHE_RESOURCE_SET res;
  if( HaArrayCount(rs->free_resource_sets) > 0 )
  {
    res = HaArrayLastAndDelete(rs->free_resource_sets);
    KheResourceSetClear(res);
  }
  else
    res = KheResourceSetMake(rs->resource_type, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceSelectSetCluster(KHE_RESOURCE_SELECT rs,                 */
/*    KHE_SOLN soln, HA_ARENA a)                                             */
/*                                                                           */
/*  Set rs, because it's new, or we've just had a success.                   */
/*                                                                           */
/*****************************************************************************/
static void KheResourceSelectDebug(KHE_RESOURCE_SELECT rs,
  int verbosity, int indent, FILE *fp);

static void KheResourceSelectSetCluster(KHE_RESOURCE_SELECT rs,
  KHE_SOLN soln, HA_ARENA a)
{
  int i, j, k, k_offset, k_tries, x, active_group_count, open_group_count;
  int over_count, max_count, under_count, minimum, maximum;
  bool allow_zero;  KHE_CONSTRAINT c;  KHE_RESOURCE_LOAD r_load;
  KHE_CLUSTER_BUSY_TIMES_CONSTRAINT cbtc;  KHE_MONITOR m;
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;  KHE_RESOURCE_SET rset;
  KHE_RESOURCE r, over_r, under_r, max_r /* , under_r1, under_r2 */;

  if( DEBUG5 )
    fprintf(stderr, "[ KheResourceSelectSetCluster(%s, soln, \"%s\")\n",
      KheResourceTypeId(rs->resource_type), rs->u.cl.constraint_str);
      /* rs->u.cl.resource_count); */

  /* free any existing resources and resource sets */
  for( i = KHE_UNDERLOADED;  i <= KHE_OVERLOADED;  i++ )
    HaArrayClear(rs->u.cl.load_class[i]);
  HaArrayAppend(rs->free_resource_sets, rs->u.cl.resource_sets, i);
  HaArrayClear(rs->u.cl.resource_sets);

  /* classify each resource of type rt */
  for( i = 0;  i < KheResourceTypeResourceCount(rs->resource_type);  i++ )
  {
    r = KheResourceTypeResource(rs->resource_type, i);
    r_load = KHE_UNDERLOADED;
    for( j = 0;  j < KheSolnResourceMonitorCount(soln, r);  j++ )
    {
      m = KheSolnResourceMonitor(soln, r, j);
      c = KheMonitorConstraint(m);
      if( KheMonitorTag(m) == KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG &&
          (HnStringContains(KheConstraintId(c), rs->u.cl.constraint_str, &x) ||
	   HnStringContains(KheConstraintName(c), rs->u.cl.constraint_str,&x)) )
      {
        cbtc = (KHE_CLUSTER_BUSY_TIMES_CONSTRAINT) c;
	if( KheClusterBusyTimesConstraintAllPositive(cbtc) )
	{
	  cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) m;
	  KheClusterBusyTimesMonitorActiveTimeGroupCount(cbtm,
	    &active_group_count, &open_group_count, &minimum, &maximum,
	    &allow_zero);
	  if( active_group_count > maximum )
	  {
	    /* r is overloaded */
	    r_load = max(r_load, KHE_OVERLOADED);
	  }
	  else if( active_group_count == maximum )
	  {
	    /* r is maximal */
	    r_load = max(r_load, KHE_MAXIMAL);
	  }
	}
      }
    }
    HaArrayAddLast(rs->u.cl.load_class[r_load], r);
    if( DEBUG5 )
      fprintf(stderr, "  adding %s to class %d\n", KheResourceId(r), r_load);
  }

  /* build the resource sets and add them to res */
  under_count = HaArrayCount(rs->u.cl.load_class[KHE_UNDERLOADED]);
  max_count = HaArrayCount(rs->u.cl.load_class[KHE_MAXIMAL]);
  over_count = HaArrayCount(rs->u.cl.load_class[KHE_OVERLOADED]);
  /* *** did not work well so I've withdrawn it for now
  if( under_count >= 2 )
  {
    ** build triples (over, under, under) **
    for( i = 0;  i < over_count;  i++ )
    {
      over_r = HaArray(rs->u.cl.load_class[KHE_OVERLOADED], i);
      j = FindIndex(rs, over_r, KHE_UNDERLOADED);
      under_r1 = HaArray(rs->u.cl.load_class[KHE_UNDERLOADED], j);
      if( j < under_count - 1 )
	under_r2 = HaArray(rs->u.cl.load_class[KHE_UNDERLOADED], j + 1);
      else
	under_r2 = HaArray(rs->u.cl.load_class[KHE_UNDERLOADED], j - 1);
      if( DEBUG5 )
	fprintf(stderr, "  triple (over %s, under %s, under %s)\n",
	  KheResourceId(over_r), KheResourceId(under_r1),
	  KheResourceId(under_r2));
      rset = KheResourceSetGet(rs, a);
      KheResourceSetAddResource(rset, over_r);
      KheResourceSetAddResource(rset, under_r1);
      KheResourceSetAddResource(rset, under_r2);
      HaArrayAddLast(rs->u.cl.resource_sets, rset);
    }
  }
  else
  *** */
  if( max_count > 0 )
  {
    /* build triples (over, max, under) */
    k_tries = min(max_count, 3);
    for( i = 0;  i < over_count;  i++ )
    {
      over_r = HaArray(rs->u.cl.load_class[KHE_OVERLOADED], i);
      for( j = 0;  j < under_count;  j++ )
      {
	under_r = HaArray(rs->u.cl.load_class[KHE_UNDERLOADED], j);
	k_offset = FindIndex(rs, over_r, KHE_MAXIMAL);
	for( k = 0;  k < k_tries;  k++ )
	{
	  max_r = HaArray(rs->u.cl.load_class[KHE_MAXIMAL],
	    (k_offset + k) % max_count);
	  if( DEBUG5 )
	    fprintf(stderr, "  triple (over %s, under %s, max %s)\n",
	      KheResourceId(over_r), KheResourceId(under_r),
	      KheResourceId(max_r));
	  rset = KheResourceSetGet(rs, a);
	  KheResourceSetAddResource(rset, over_r);
	  KheResourceSetAddResource(rset, under_r);
	  KheResourceSetAddResource(rset, max_r);
	  HaArrayAddLast(rs->u.cl.resource_sets, rset);
	}
      }
    }
  }
  else
  {
    /* no maximal resources, so just build pairs (over, under) */
    for( i = 0;  i < over_count; i++ )
    {
      over_r = HaArray(rs->u.cl.load_class[KHE_OVERLOADED], i);
      for( j = 0; j < under_count; j++ )
      {
	under_r = HaArray(rs->u.cl.load_class[KHE_UNDERLOADED], j);
	if( DEBUG5 )
	  fprintf(stderr, "  pair (over %s, under %s)\n",
	    KheResourceId(over_r), KheResourceId(under_r));
	rset = KheResourceSetGet(rs, a);
	KheResourceSetAddResource(rset, over_r);
	KheResourceSetAddResource(rset, under_r);
	HaArrayAddLast(rs->u.cl.resource_sets, rset);
      }
    }
  }

  if( DEBUG5 )
  {
    fprintf(stderr, "  KheResourceSelectSetCluster made resource sets:\n");
    KheResourceSelectDebug(rs, 2, 2, stderr);
    fprintf(stderr, "] KheResourceSelectSetCluster returning\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SELECT KheResourceSelectDoCluster(KHE_RESOURCE_TYPE rt,     */
/*    KHE_SOLN soln, char *constraint_str, int resource_count, HA_ARENA a)   */
/*                                                                           */
/*  Make and return a cluster resource select with these attributes.         */
/*                                                                           */
/*****************************************************************************/

/* ***
static void KheResourceSelectDebug(KHE_RESOURCE_SELECT rs,
  int verbosity, int indent, FILE *fp);

static KHE_RESOURCE_SELECT KheResourceSelectDoCluster(KHE_RESOURCE_TYPE rt,
  KHE_SOLN soln, char *constraint_str, int resource_count, HA_ARENA a)
{
  if( DEBUG5 )
    fprintf(stderr, "[ KheResourceSelectDoCluster(%s, soln, \"%s\", %d)\n",
      KheResourceTypeId(rt), constraint_str, resource_count);

  res = KheResourceSelectMakeCluster(rt, constraint_str, resource_count, a);
  KheResourceSelectSetCluster(res, soln, a);

  if( DEBUG5 )
  {
    fprintf(stderr, "  KheResourceSelectMakeCluster made resource sets:\n");
    KheResourceSelectDebug(res, 2, 2, stderr);
    fprintf(stderr, "] KheResourceSelectMakeCluster returning\n");
  }
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SELECT KheResourceSelectParse(char **str,                   */
/*    KHE_RESOURCE_TYPE rt, KHE_SOLN soln, bool allow_multiple, HA_ARENA a)  */
/*                                                                           */
/*  Parse a resource select.  If allow_multiple is true, the syntax is       */
/*                                                                           */
/*    resource_select  ::=  choose  |  cluster  |  literal                   */
/*    choose           ::=  "choose" "(" int ")"                             */
/*    cluster          ::=  "cluster" "(" str "," int ")"                    */
/*    literal          ::=  int_set { ";" int_set }                          */
/*    int_set          ::=  int { "," int }                                  */
/*                                                                           */
/*  If allow_multiple is false, the { ";" int_set } part is omitted.         */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SELECT KheResourceSelectParse(char **str,
  KHE_RESOURCE_TYPE rt, KHE_SOLN soln, bool allow_multiple, HA_ARENA a)
{
  KHE_RESOURCE_SELECT res;  KHE_RESOURCE_SET rs;
  int choose_len /* , resource_count */;  char *constraint_str;
  if( KheStrSkip(str, "choose") )
  {
    /* "choose" "(" int ")" */
    KheStrSkip(str, "(");
    choose_len = KheIntParse(str);
    HnAssert(choose_len >= 1, "KheResourceSelectParse: choose_len <= 0");
    res = KheResourceSelectMakeChoose(rt, choose_len, a);
    KheStrSkip(str, ")");
  }
  else if( KheStrSkip(str, "cluster") )
  {
    /* "cluster" "(" str "," int ")" */
    KheStrSkip(str, "(");
    constraint_str = KheStringParse(str, ')');
    /* ***
    constraint_str = KheStringParse(str, ',');
    KheStrSkip(str, ",");
    resource_count = KheIntParse(str);
    *** */
    KheStrSkip(str, ")");
    /*  ***
    res = KheResourceSelectDoCluster(rt, soln, constraint_str,resource_count,a);
    *** */
    res = KheResourceSelectMakeCluster(rt, constraint_str,
      /* resource_count, */ a);
    KheResourceSelectSetCluster(res, soln, a);
  }
  else
  {
    res = KheResourceSelectMakeEnumerated(rt, a);
    do
    {
      rs = KheResourceSetParse(str, rt, a);
      HaArrayAddLast(res->u.en.resource_sets, rs);
    } while( allow_multiple && KheStrSkip(str, ";") );
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SET KheResourceSelectChooseRandom(KHE_RESOURCE_SELECT rs,   */
/*    KHE_RANDOM_GENERATOR *rgen)                                            */
/*                                                                           */
/*  Return rs's scratch rset, containing rs->choose_len resources chosen     */
/*  at random.                                                               */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SET KheResourceSelectChooseRandom(KHE_RESOURCE_SELECT rs,
  KHE_RANDOM_GENERATOR *rgen)
{
  KHE_RESOURCE_SET rset;  KHE_RESOURCE tmp_r;  int i, pos;
  HnAssert(rs->kind == KHE_RESOURCE_SELECT_CHOOSE,
    "KheResourceSelectChooseRandom internal error");
  rset = rs->u.ch.scratch_resource_set;
  KheResourceSetClear(rset);
  for( i = 0;  i < rs->u.ch.choose_len;  i++ )
  {
    pos = KheRandomGeneratorNextRange(rgen, i,
      HaArrayCount(rs->u.ch.randomized_resources) - 1);
    HaArraySwap(rs->u.ch.randomized_resources, i, pos, tmp_r);
    KheResourceSetAddResource(rset, HaArray(rs->u.ch.randomized_resources, i));
  }
  return rset;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheResourceSelectIsOpen(KHE_RESOURCE_SELECT rs,                     */
/*    KHE_RANDOM_GENERATOR *rgen, ARRAY_KHE_RESOURCE *rset, int i)           */
/*                                                                           */
/*  If rs is open to iteration i, return true and set *rset to the resource  */
/*  set to use on that iteration.                                            */
/*                                                                           */
/*****************************************************************************/

static bool KheResourceSelectIsOpen(KHE_RESOURCE_SELECT rs, 
  KHE_RANDOM_GENERATOR *rgen, KHE_RESOURCE_SET *rset, int i)
{
  switch( rs->kind )
  {
    case KHE_RESOURCE_SELECT_ENUMERATED:

      if( i < HaArrayCount(rs->u.en.resource_sets) )
      {
	/* return the ith resource set */
	return *rset = HaArray(rs->u.en.resource_sets, i), true;
      }
      break;

    case KHE_RESOURCE_SELECT_CHOOSE:

      if( rs->u.ch.choose_len < HaArrayCount(rs->u.ch.randomized_resources)
	  || i == 0 )
      {
	/* return rs->choose_len resources chosen at random, but not */
	/* if there is only one choice and it has already been made */
	return *rset = KheResourceSelectChooseRandom(rs, rgen), true;
      }
      break;

    case KHE_RESOURCE_SELECT_CLUSTER:

      if( i < HaArrayCount(rs->u.cl.resource_sets) )
      {
	/* return the ith resource set */
	return *rset = HaArray(rs->u.cl.resource_sets, i), true;
      }
      break;

    default:

      HnAbort("KheResourceSelectIsOpen internal error");
      break;
  }

  /* no luck, return false */
  return *rset = NULL, false;
}


/*****************************************************************************/
/*                                                                           */
/*  KheResourceSelectForEach(KHE_RESOURCE_SELECT rs,                         */
/*    KHE_RANDOM_GENERATOR *rgen, KHE_RESOURCE_SET rset, int i)              */
/*                                                                           */
/*****************************************************************************/

#define KheResourceSelectForEach(rs, rgen, rset, i)		\
for( (i) = 0; KheResourceSelectIsOpen((rs), (rgen), &(rset), (i)); (i)++ )


/*****************************************************************************/
/*                                                                           */
/*  KHE_RESOURCE_SET KheResourceSelectFirstResourceSet(                      */
/*    KHE_RESOURCE_SELECT rs, KHE_RANDOM_GENERATOR *rgen)                    */
/*                                                                           */
/*  Return the first resource set of rs.  Generate it randomly if needed.    */
/*                                                                           */
/*****************************************************************************/

static KHE_RESOURCE_SET KheResourceSelectFirstResourceSet(
  KHE_RESOURCE_SELECT rs, KHE_RANDOM_GENERATOR *rgen)
{
  switch( rs->kind )
  {
    case KHE_RESOURCE_SELECT_ENUMERATED:

      if( HaArrayCount(rs->u.en.resource_sets) > 0 )
      {
	/* choose the first resource set */
	return HaArrayFirst(rs->u.en.resource_sets);
      }
      break;

    case KHE_RESOURCE_SELECT_CHOOSE:

      if( rs->u.ch.scratch_resource_set == NULL )
	rs->u.ch.scratch_resource_set = KheResourceSelectChooseRandom(rs, rgen);
      return rs->u.ch.scratch_resource_set;

    case KHE_RESOURCE_SELECT_CLUSTER:

      if( HaArrayCount(rs->u.cl.resource_sets) > 0 )
      {
	/* choose the first resource set */
	return HaArrayFirst(rs->u.cl.resource_sets);
      }
      break;

    default:

      HnAbort("KheResourceSelectFirstResourceSet internal error");
      break;
  }

  /* not open */
  HnAbort("KheResourceSelectFirstResourceSet: no first resource set");
  return NULL;  /* keep compiler happy */
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceSelectAddFirstResourceSetDisplay(KHE_RESOURCE_SELECT rs, */
/*    HA_ARRAY_NCHAR *res)                                                   */
/*                                                                           */
/*  Add a representation of rs's first element to res.                       */
/*                                                                           */
/*****************************************************************************/

static void KheResourceSelectAddFirstResourceSetDisplay(KHE_RESOURCE_SELECT rs,
  HA_ARRAY_NCHAR *res)
{
  KHE_RESOURCE_SET rset;
  switch( rs->kind )
  {
    case KHE_RESOURCE_SELECT_ENUMERATED:

      HnAssert(HaArrayCount(rs->u.en.resource_sets) > 0,
	"KheResourceSelectAddFirstResourceSetDisplay internal error");
      rset = HaArrayFirst(rs->u.en.resource_sets);
      KheResourceSetAddDisplay(rset, res);
      break;

    case KHE_RESOURCE_SELECT_CHOOSE:

      HnStringAdd(res, "%d %s resources", rs->u.ch.choose_len,
	KheResourceTypeId(rs->resource_type));
      break;

    case KHE_RESOURCE_SELECT_CLUSTER:

      HnAssert(HaArrayCount(rs->u.cl.resource_sets) > 0,
	"KheResourceSelectAddFirstResourceSetDisplay internal error");
      rset = HaArrayFirst(rs->u.cl.resource_sets);
      KheResourceSetAddDisplay(rset, res);
      break;

    default:

      HnAbort("KheResourceSelectAddFirstResourceSetDisplay internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheResourceSelectDebug(KHE_RESOURCE_SELECT rs,                      */
/*    int verbosity, int indent, FILE *fp)                                   */
/*                                                                           */
/*  Debug print of rs onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheResourceSelectDebug(KHE_RESOURCE_SELECT rs,
  int verbosity, int indent, FILE *fp)
{
  KHE_RESOURCE_SET rset;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  switch( rs->kind )
  {
    case KHE_RESOURCE_SELECT_ENUMERATED:

      HaArrayForEach(rs->u.en.resource_sets, rset, i)
      {
	if( i > 0 )
	{
	  if( indent > 0 )
	    fprintf(fp, "\n%*s", indent, "");
	  else
	    fprintf(fp, ";");
	}
	KheResourceSetDebug(rset, verbosity, -1, fp);
      }
      break;

    case KHE_RESOURCE_SELECT_CHOOSE:

      fprintf(fp, "choose(%d)", rs->u.ch.choose_len);
      break;

    case KHE_RESOURCE_SELECT_CLUSTER:

      HaArrayForEach(rs->u.cl.resource_sets, rset, i)
      {
	if( i > 0 )
	{
	  if( indent > 0 )
	    fprintf(fp, "\n%*s", indent, "");
	  else
	    fprintf(fp, ";");
	}
	KheResourceSetDebug(rset, verbosity, -1, fp);
      }
      break;

    default:

      break;
  }
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_INTERVAL"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheIntervalMake(int first, int last)                        */
/*                                                                           */
/*  Make a new interval object with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

/* *** moved to khe_sm_interval.c
static KHE_INTERVAL KheIntervalMake(int first, int last)
{
  KHE_INTERVAL res;
  res.first = first;
  res.last = last;
  return res;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheIntervalParse(char **str)                                */
/*                                                                           */
/*  Parse an interval.                                                       */
/*                                                                           */
/*****************************************************************************/

static KHE_INTERVAL KheIntervalParse(char **str)
{
  int first, last;
  first = KheIntParse(str);
  if( !KheStrSkip(str, "-") )
    HnAbort("error where interval hyphen expected in rs_drs_test option");
  last = KheIntParse(str);
  return KheIntervalMake(first, last);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheIntervalEqual(KHE_INTERVAL in1, KHE_INTERVAL in2)                */
/*                                                                           */
/*  Return true if in1 and in2 are equal.                                    */
/*                                                                           */
/*****************************************************************************/

/* *** moved to khe_sm_interval.c
static bool KheIntervalEqual(KHE_INTERVAL in1, KHE_INTERVAL in2)
{
  return in1.first == in2.first && in1.last == in2.last;
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheIntervalAddDisplay(KHE_INTERVAL in, HA_ARRAY_NCHAR *res)         */
/*                                                                           */
/*  Add a display of in to res.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheIntervalAddDisplay(KHE_INTERVAL in, HA_ARRAY_NCHAR *res)
{
  HnStringAdd(res, "%d-%d", KheIntervalFirst(in), KheIntervalLast(in));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheIntervalDebug(KHE_INTERVAL in, FILE *fp)                         */
/*                                                                           */
/*  Debug print of in onto fp.                                               */
/*                                                                           */
/*****************************************************************************/

static void KheIntervalDebug(KHE_INTERVAL in, FILE *fp)
{
  fprintf(fp, "%d-%d", KheIntervalFirst(in), KheIntervalLast(in));
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_INTERVAL_SET"                                             */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL_SET KheIntervalSetMake(HA_ARENA a)                          */
/*                                                                           */
/*  Make a new, empty interval set object.                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_INTERVAL_SET KheIntervalSetMake(HA_ARENA a)
{
  KHE_INTERVAL_SET res;
  HaMake(res, a);
  HaArrayInit(res->intervals, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL_SET KheIntervalSetParse(char **str, HA_ARENA a)             */
/*                                                                           */
/*  Parse an interval set.                                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_INTERVAL_SET KheIntervalSetParse(char **str, HA_ARENA a)
{
  KHE_INTERVAL in;  KHE_INTERVAL_SET res;
  res = KheIntervalSetMake(a);
  do
  {
    in = KheIntervalParse(str);
    HaArrayAddLast(res->intervals, in);
  } while( KheStrSkip(str, ",") );
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheIntervalSetEqual(KHE_INTERVAL_SET is1, KHE_INTERVAL_SET is2)     */
/*                                                                           */
/*  Return true if is1 and is2 are equal.                                    */
/*                                                                           */
/*****************************************************************************/

static bool KheIntervalSetEqual(KHE_INTERVAL_SET is1, KHE_INTERVAL_SET is2)
{
  KHE_INTERVAL i1, i2;  int i;
  if( is1 == is2 )
    return true;
  if( HaArrayCount(is1->intervals) != HaArrayCount(is2->intervals) )
    return false;
  for( i = 0;  i < HaArrayCount(is1->intervals);  i++ )
  {
    i1 = HaArray(is1->intervals, i);
    i2 = HaArray(is2->intervals, i);
    if( !KheIntervalEqual(i1, i2) )
      return false;
  }
  return true;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheIntervalSetAddDisplay(KHE_INTERVAL_SET is, HA_ARRAY_NCHAR *res)  */
/*                                                                           */
/*  Add a representation of is to res.                                       */
/*                                                                           */
/*****************************************************************************/

static void KheIntervalSetAddDisplay(KHE_INTERVAL_SET is, HA_ARRAY_NCHAR *res)
{
  KHE_INTERVAL in;  int i;
  HaArrayForEach(is->intervals, in, i)
  {
    if( i > 0 )
      HnStringAdd(res, ",");
    KheIntervalAddDisplay(in, res);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheIntervalSetDebug(KHE_INTERVAL_SET is, FILE *fp)                  */
/*                                                                           */
/*  Debug printf of is onto fp.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheIntervalSetDebug(KHE_INTERVAL_SET is, FILE *fp)
{
  KHE_INTERVAL in;  int i;
  HaArrayForEach(is->intervals, in, i)
  {
    if( i > 0 )
      fprintf(fp, ",");
    KheIntervalDebug(in, fp);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DAY_SELECT"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectMake(KHE_FRAME days_frame, int choose_len,    */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Make a new day select object with these attributes.                      */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectMake(KHE_FRAME days_frame, int choose_len,
  HA_ARENA a)
{
  KHE_DAY_SELECT res;
  HaMake(res, a);
  res->days_frame = days_frame;
  res->choose_len = choose_len;
  HaArrayInit(res->interval_sets, a);
  res->single_interval_set = NULL;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectIntervalSetsParse(char **str,                 */
/*    KHE_FRAME days_frame, bool allow_multiple, HA_ARENA a)                 */
/*                                                                           */
/*  Parse a day select object which is a sequence of interval sets.          */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectIntervalSetsParse(char **str,
  KHE_FRAME days_frame, bool allow_multiple, HA_ARENA a)
{
  KHE_DAY_SELECT res;  KHE_INTERVAL_SET is;
  res = KheDaySelectMake(days_frame, 0, a);
  do
  {
    is = KheIntervalSetParse(str, a);
    HaArrayAddLast(res->interval_sets, is);
  } while( allow_multiple && KheStrSkip(str, ";") );
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectChooseMake(KHE_FRAME days_frame,              */
/*    int choose_len, HA_ARENA a)                                            */
/*                                                                           */
/*  Make a new "choose(x)" day select object.                                */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectChooseMake(KHE_FRAME days_frame,
  int choose_len, HA_ARENA a)
{
  KHE_DAY_SELECT res;  KHE_INTERVAL in;  KHE_INTERVAL_SET is;
  res = KheDaySelectMake(days_frame, choose_len, a);
  in = KheIntervalMake(0, 0);  /* placeholder, will be reset later */
  is = KheIntervalSetMake(a);
  HaArrayAddLast(is->intervals, in);
  HaArrayAddLast(res->interval_sets, is);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectChooseParse(char **str,                       */
/*    KHE_FRAME days_frame, HA_ARENA a)                                      */
/*                                                                           */
/*  Parse a "choose(x)" day select; the "choose" has already been read.      */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectChooseParse(char **str,
  KHE_FRAME days_frame, HA_ARENA a)
{
  int choose_len;
  KheStrSkip(str, "(");
  choose_len = KheIntParse(str);
  if( choose_len > KheFrameTimeGroupCount(days_frame) )
    choose_len = KheFrameTimeGroupCount(days_frame);
  KheStrSkip(str, ")");
  return KheDaySelectChooseMake(days_frame, choose_len, a);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectAllMake(KHE_FRAME days_frame,                 */
/*    int len, int inc, HA_ARENA a)                                          */
/*                                                                           */
/*  Make an "all(len, inc)" day select object, by expanding into an          */
/*  explicit interval set.                                                   */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectAllMake(KHE_FRAME days_frame,
  int len, int inc, HA_ARENA a)
{
  KHE_INTERVAL in;  KHE_INTERVAL_SET is;  KHE_DAY_SELECT res;  int i, last;
  res = KheDaySelectMake(days_frame, -1, a);
  for( i = 0;  i < KheFrameTimeGroupCount(days_frame);  i += inc )
  {
    last = i + len - 1;
    if( last >= KheFrameTimeGroupCount(days_frame) )
      last = KheFrameTimeGroupCount(days_frame) - 1;
    in = KheIntervalMake(i, last);
    is = KheIntervalSetMake(a);
    HaArrayAddLast(is->intervals, in);
    HaArrayAddLast(res->interval_sets, is);
    if( last == KheFrameTimeGroupCount(days_frame) - 1 )
      break;
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectAllParse(char **str, KHE_FRAME days_frame,    */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Parse "all", "all(len)", or "all(len, inc)"; "all" is already parsed.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectAllParse(char **str, KHE_FRAME days_frame,
  HA_ARENA a)
{
  int len, inc;

  /* get len and inc from all, all(len), or all(len, inc) */
  if( KheStrSkip(str, "(") )
  {
    len = KheIntParse(str);
    if( KheStrSkip(str, ",") )
      inc = KheIntParse(str);
    else
      inc = len;
    KheStrSkip(str, ")");
  }
  else
    len = inc = KheFrameTimeGroupCount(days_frame);

  /* sanity check on len */
  if( len < 0 )
    len = 0;
  else if( len > KheFrameTimeGroupCount(days_frame) )
    len = KheFrameTimeGroupCount(days_frame);

  /* sanity check on inc */
  if( inc <= 0 )
    inc = 1;

  /* build and return the object */
  return KheDaySelectAllMake(days_frame, len, inc, a);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_DAY_SELECT KheDaySelectParse(char **str, KHE_FRAME days_frame,       */
/*    bool allow_multiple, HA_ARENA a)                                       */
/*                                                                           */
/*  Parse a day select object.  If allow_multiple is true, the syntax is     */
/*                                                                           */
/*    day_select     ::=  interval_sets  |  all  |  choose                   */
/*                                                                           */
/*    interval_sets  ::=  interval_set { ";" interval_set }                  */
/*    interval_set   ::=  interval { "," interval }                          */
/*    interval       ::=  int "-" int                                        */
/*                                                                           */
/*    all            ::=  "all" [ "(" int [ "," int ] ")" ]                  */
/*                                                                           */
/*    choose         ::=  "choose" "(" int ")"                               */
/*                                                                           */
/*  If allow_multiple is false, the { ";" interval_set } part is omitted.    */
/*                                                                           */
/*****************************************************************************/

static KHE_DAY_SELECT KheDaySelectParse(char **str, KHE_FRAME days_frame,
  bool allow_multiple, HA_ARENA a)
{
  if( is_digit(**str) )
    return KheDaySelectIntervalSetsParse(str, days_frame, allow_multiple, a);
  else if( KheStrSkip(str, "all") )
    return KheDaySelectAllParse(str, days_frame, a);
  else if( KheStrSkip(str, "choose") )
    return KheDaySelectChooseParse(str, days_frame, a);
  else
  {
    HnAbort("KheDynamicResourceVLSNSolve: error in rs_drs_days option");
    return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL_SET KheDaySelectChooseRandom(KHE_DAY_SELECT ds,             */
/*    KHE_RANDOM_GENERATOR *rgen)                                            */
/*                                                                           */
/*  Choose an interval set for ds at random.                                 */
/*                                                                           */
/*****************************************************************************/

static KHE_INTERVAL_SET KheDaySelectChooseRandom(KHE_DAY_SELECT ds,
  KHE_RANDOM_GENERATOR *rgen)
{
  int tg_count, len, start_pos;  KHE_INTERVAL_SET is;  KHE_INTERVAL in;

  /* find start_pos */
  tg_count = KheFrameTimeGroupCount(ds->days_frame);
  len = ds->choose_len;
  start_pos = KheRandomGeneratorNextRange(rgen, 0, tg_count - len);

  /* add [start_pos, start_pos + len - 1] to interval set */
  HnAssert(HaArrayCount(ds->interval_sets) == 1,
    "KheDaySelectChooseRandom internal error 1");
  is = HaArrayFirst(ds->interval_sets);
  HnAssert(HaArrayCount(is->intervals) == 1,
    "KheDaySelectChooseRandom internal error 2");
  in.first = start_pos;
  in.last = start_pos + len - 1;
  HaArrayPut(is->intervals, 0, in);
  if( DEBUG4 )
    fprintf(stderr, "KheDaySelectChooseRandom(range 0-%d) returning %d-%d\n",
      tg_count - len, in.first, in.last);
  return is;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheDaySelectIsOpen(KHE_DAY_SELECT ds,                               */
/*    KHE_RANDOM_GENERATOR *rgen, KHE_INTERVAL_SET *is, int i)               */
/*                                                                           */
/*  If there is an i'th interval set in ds, set *is to it and return true.   */
/*  Otherwise return false.                                                  */
/*                                                                           */
/*  When choosing at random, there is just one (random) interval set.        */
/*                                                                           */
/*****************************************************************************/

static bool KheDaySelectIsOpen(KHE_DAY_SELECT ds,
  KHE_RANDOM_GENERATOR *rgen, KHE_INTERVAL_SET *is, int i)
{
  if( ds->choose_len >= 1 && i == 0 )
  {
    /* return an interval set containing one interval chosen at random */
    return *is = KheDaySelectChooseRandom(ds, rgen), true;
  }
  else if( i < HaArrayCount(ds->interval_sets) )
  {
    /* return the ith interval set */
    return *is = HaArray(ds->interval_sets, i), true;
  }
  else
  {
    /* not open */
    return *is = NULL, false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KheDaySelectForEach(KHE_DAY_SELECT ds, KHE_INTERVAL_SET is, int i)       */
/*                                                                           */
/*****************************************************************************/

#define KheDaySelectForEach(ds, rgen, iset, i)				\
for( (i) = 0;  KheDaySelectIsOpen((ds), (rgen), &(iset), (i));  (i)++ )


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL_SET KheDaySelectFirstIntervalSet(KHE_DAY_SELECT ds,         */
/*    KHE_RANDOM_GENERATOR *rgen)                                            */
/*                                                                           */
/*  Return the first interval set of ds.  Generate it randomly if needed.    */
/*                                                                           */
/*****************************************************************************/

static KHE_INTERVAL_SET KheDaySelectFirstIntervalSet(KHE_DAY_SELECT ds,
  KHE_RANDOM_GENERATOR *rgen)
{
  if( ds->choose_len > 0 )
  {
    /* choose an interval set at random, but only do it once */
    if( ds->single_interval_set == NULL )
      ds->single_interval_set = KheDaySelectChooseRandom(ds, rgen);
    return ds->single_interval_set;
  }
  else if( HaArrayCount(ds->interval_sets) > 0 )
  {
    /* choose the first interval set */
    return HaArrayFirst(ds->interval_sets);
  }
  else
  {
    /* not open */
    HnAbort("KheDaySelectFirstIntervalSet: no first interval set");
    return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDaySelectAddFirstIntervalSetDisplay(KHE_DAY_SELECT ds,           */
/*    HA_ARRAY_NCHAR *res)                                                   */
/*                                                                           */
/*  Add a display of ds's first element to res.                              */
/*                                                                           */
/*****************************************************************************/

static void KheDaySelectAddFirstIntervalSetDisplay(KHE_DAY_SELECT ds,
  HA_ARRAY_NCHAR *res)
{
  KHE_INTERVAL_SET is;
  if( ds->choose_len > 0 )
    HnStringAdd(res, "%d-day interval", ds->choose_len);
  else
  {
    HnAssert(HaArrayCount(ds->interval_sets) > 0,
      "KheDaySelectAddFirstIntervalSetDisplay internal error");
    is = HaArrayFirst(ds->interval_sets);
    KheIntervalSetAddDisplay(is, res);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDaySelectDebug(KHE_DAY_SELECT ds, int verbosity,                 */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of ds onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheDaySelectDebug(KHE_DAY_SELECT ds, int verbosity,
  int indent, FILE *fp)
{
  KHE_INTERVAL_SET is;  int i;
  if( indent > 0 )
    fprintf(fp, "%*s", indent, "");
  if( ds->choose_len > 0 )
    fprintf(fp, "choose(%d)", ds->choose_len);
  else
  {
    HaArrayForEach(ds->interval_sets, is, i)
    {
      if( i > 0 )
	fprintf(fp, ";");
      KheIntervalSetDebug(is, fp);
    }
  }
  if( indent >= 0 )
    fprintf(fp, "\n");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DOM_KIND"                                                 */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DRS_DOM_KIND KheDomKindParse(char **str)                             */
/*                                                                           */
/*  Read the next dom kind and return it.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_DRS_DOM_KIND KheDomKindParse(char **str)
{
  if( KheStrSkip(str, "list_none") )
    return KHE_DRS_DOM_LIST_NONE;
  else if( KheStrSkip(str, "list_strong") )
    return KHE_DRS_DOM_LIST_STRONG;
  else if( KheStrSkip(str, "list_tradeoff") )
    return KHE_DRS_DOM_LIST_TRADEOFF;
  else if( KheStrSkip(str, "list_uniform") )
    return KHE_DRS_DOM_LIST_UNIFORM;
  else if( KheStrSkip(str, "hash_weak") )
    return KHE_DRS_DOM_HASH_WEAK;
  else if( KheStrSkip(str, "hash_medium") )
    return KHE_DRS_DOM_HASH_MEDIUM;
  else if( KheStrSkip(str, "trie_strong") )
    return KHE_DRS_DOM_TRIE_STRONG;
  else if( KheStrSkip(str, "trie_tradeoff") )
    return KHE_DRS_DOM_TRIE_TRADEOFF;
  else if( KheStrSkip(str, "indexed_tradeoff") )
    return KHE_DRS_DOM_INDEXED_TRADEOFF;
  else if( KheStrSkip(str, "indexed_uniform") )
    return KHE_DRS_DOM_INDEXED_UNIFORM;
  else
  {
    HnAbort("error where dom kind name expected in rs_drs_test option");
    return KHE_DRS_DOM_LIST_NONE;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDomKindName(KHE_DRS_DOM_KIND dom_kind)                          */
/*                                                                           */
/*  Return the string name of dom_kind.                                      */
/*                                                                           */
/*****************************************************************************/

static char *KheDomKindName(KHE_DRS_DOM_KIND dom_kind)
{
  switch( dom_kind )
  {
    case KHE_DRS_DOM_LIST_NONE:		return "list_none";
    case KHE_DRS_DOM_LIST_STRONG:	return "list_strong";
    case KHE_DRS_DOM_LIST_TRADEOFF:	return "list_tradeoff";
    case KHE_DRS_DOM_LIST_UNIFORM:	return "list_uniform";
    case KHE_DRS_DOM_HASH_WEAK:		return "hash_weak";
    case KHE_DRS_DOM_HASH_MEDIUM:	return "hash_medium";
    case KHE_DRS_DOM_TRIE_STRONG:	return "trie_strong";
    case KHE_DRS_DOM_TRIE_TRADEOFF:	return "trie_tradeoff";
    case KHE_DRS_DOM_INDEXED_TRADEOFF:	return "indexed_tradeoff";
    case KHE_DRS_DOM_INDEXED_UNIFORM:	return "indexed_uniform";

    default:
      HnAbort("KheDomKindName internal error");
      return NULL;  /* keep compiler happy */
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_SOLVE_ARGUMENTS"                                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLVE_ARGUMENTS KheSolveArgumentsGet(KHE_OPTIONS options, HA_ARENA a)*/
/*                                                                           */
/*  Get a set of solve arguments from options.                               */
/*                                                                           */
/*****************************************************************************/

static KHE_SOLVE_ARGUMENTS KheSolveArgumentsGet(KHE_SOLN soln,
  KHE_RESOURCE_TYPE rt, KHE_FRAME days_frame, KHE_OPTIONS options,
  HA_ARENA a)
{
  KHE_SOLVE_ARGUMENTS res;  char *str;

  /* make a new uninitialized solve arguments object */
  HaMake(res, a);

  /* get the resources option */
  str = KheOptionsGet(options, "rs_drs_resources", "choose(3)");
  res->resources = KheResourceSelectParse(&str, rt, soln, true, a);

  /* get the days option */
  str = KheOptionsGet(options, "rs_drs_days", "choose(28)");
  res->days = KheDaySelectParse(&str, days_frame, true, a);

  /* get the Boolean options */
  res->priqueue =
    KheOptionsGetBool(options, "rs_drs_priqueue", false);
  res->extra_selection =
    KheOptionsGetBool(options, "rs_drs_extra_selection", true);
  res->expand_by_shifts =
    KheOptionsGetBool(options, "rs_drs_expand_by_shifts", true);
  res->shift_pairs =
    KheOptionsGetBool(options, "rs_drs_shift_pairs", false);
  res->correlated_exprs =
    KheOptionsGetBool(options, "rs_drs_correlated_exprs", false);

  /* get the int options */
  res->daily_expand_limit =
    KheOptionsGetInt(options, "rs_drs_daily_expand_limit", 10000);
  res->daily_prune_trigger =
    KheOptionsGetInt(options, "rs_drs_daily_prune_trigger", 0);
  res->resource_expand_limit =
    KheOptionsGetInt(options, "rs_drs_resource_expand_limit", 0);
  res->dom_approx =
    KheOptionsGetInt(options, "rs_drs_dom_approx", 0);

  /* get the cache options */
  str = KheOptionsGet(options, "rs_drs_dom_kind", "indexed_uniform");
  res->main_dom_kind = KheDomKindParse(&str);
  str = KheOptionsGet(options, "rs_drs_cache_dom_kind", "nocache");
  if( strcmp(str, "nocache") == 0 )
    res->cache = false, res->cache_dom_kind = KHE_DRS_DOM_LIST_NONE;
  else if( strcmp(str, "same") == 0 )
    res->cache = true, res->cache_dom_kind = res->main_dom_kind;
  else
    res->cache = true, res->cache_dom_kind = KheDomKindParse(&str);

  /* all done */
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLVE_ARGUMENTS KheSolveArgumentsCopy(KHE_SOLVE_ARGUMENTS sa,        */
/*    HA_ARENA a)                                                            */
/*                                                                           */
/*  Make a copy of sa.                                                       */
/*                                                                           */
/*  Implementation note.  There is no need for a deep copy of sa->resources  */
/*  and sa->days, because these values are either used as is or changed to   */
/*  a completely different value.  In fact, a deep copy would be wrong,      */
/*  because it would change the meaning of choose(x).                        */
/*                                                                           */
/*****************************************************************************/

static KHE_SOLVE_ARGUMENTS KheSolveArgumentsCopy(KHE_SOLVE_ARGUMENTS sa,
  HA_ARENA a)
{
  KHE_SOLVE_ARGUMENTS res;
  HaMake(res, a);
  res->resources = sa->resources;
  res->days = sa->days;
  res->priqueue = sa->priqueue;
  res->extra_selection = sa->extra_selection;
  res->expand_by_shifts = sa->expand_by_shifts;
  res->shift_pairs = sa->shift_pairs;
  res->correlated_exprs = sa->correlated_exprs;
  res->daily_expand_limit = sa->daily_expand_limit;
  res->daily_prune_trigger = sa->daily_prune_trigger;
  res->resource_expand_limit = sa->resource_expand_limit;
  res->dom_approx = sa->dom_approx;
  res->main_dom_kind = sa->main_dom_kind;
  res->cache = sa->cache;
  res->cache_dom_kind = sa->cache_dom_kind;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSolveArgumentsUpdateWithDimPosOpt(KHE_SOLVE_ARGUMENTS sa,        */
/*    KHE_DIM_POS_OPT opt)                                                   */
/*                                                                           */
/*  Update sa to reflect the option in opt.                                  */
/*                                                                           */
/*****************************************************************************/

static void KheSolveArgumentsUpdateWithDimPosOpt(KHE_SOLVE_ARGUMENTS sa,
  KHE_DIM_POS_OPT opt)
{
  switch( opt->tag )
  {
    case 'R':

      sa->resources = opt->u.resource_select_val;
      break;

    case 'D':

      sa->days = opt->u.day_select_val;
      break;

    case 'Q':

      sa->priqueue = opt->u.bool_val;
      break;

    case 'E':

      sa->extra_selection = opt->u.bool_val;
      break;

    case 'S':

      sa->expand_by_shifts = opt->u.bool_val;
      break;

    case 'P':

      sa->shift_pairs = opt->u.bool_val;
      break;

    case 'C':

      sa->correlated_exprs = opt->u.bool_val;
      break;

    case 'L':

      sa->daily_expand_limit = opt->u.int_val;
      break;

    case 'T':

      sa->daily_prune_trigger = opt->u.int_val;
      break;

    case 'M':

      sa->resource_expand_limit = opt->u.int_val;
      break;

    case 'A':

      sa->dom_approx = opt->u.int_val;
      break;

    case 'K':

      sa->main_dom_kind = opt->u.dom_kind_val;
      break;

    case 'J':

      sa->cache = true;
      sa->cache_dom_kind = opt->u.dom_kind_val;
      break;

    default:

      HnAbort("KheSolveArgumentsUpdateWithDimPosOpt internal error");
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_SOLVE_ARGUMENTS KheSolveArgumentsCopyAndUpdate(                      */
/*    KHE_SOLVE_ARGUMENTS sa, KHE_DIM_POS dim_pos, HA_ARENA a)               */
/*                                                                           */
/*  Copy sa and update it to reflect dim_pos.                                */
/*                                                                           */
/*****************************************************************************/

static KHE_SOLVE_ARGUMENTS KheSolveArgumentsCopyAndUpdate(
  KHE_SOLVE_ARGUMENTS sa, KHE_DIM_POS dim_pos, HA_ARENA a)
{
  KHE_DIM_POS_OPT opt;  int i;
  KHE_SOLVE_ARGUMENTS res;
  res = KheSolveArgumentsCopy(sa, a);
  HaArrayForEach(dim_pos->options, opt, i)
    KheSolveArgumentsUpdateWithDimPosOpt(res, opt);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheSolveArgumentsDebug(KHE_SOLVE_ARGUMENTS sa, int verbosity,       */
/*    int indent, FILE *fp)                                                  */
/*                                                                           */
/*  Debug print of sa onto fp with the given verbosity and indent.           */
/*                                                                           */
/*****************************************************************************/

static void KheBoolDebug(char *name, bool val, int verbosity,
  int indent, FILE *fp)
{
  fprintf(fp, "%*s%s %s\n", indent, "", name, bool_show(val));
}

static void KheIntDebug(char *name, int val, int verbosity,
  int indent, FILE *fp)
{
  fprintf(fp, "%*s%s %d\n", indent, "", name, val);
}

static void KheSolveArgumentsDebug(KHE_SOLVE_ARGUMENTS sa, int verbosity,
  int indent, FILE *fp)
{
  if( indent >= 0 )
  {
    fprintf(stderr, "%*s[ KHE_SOLVE_ARGUMENTS\n", indent, "");
    KheResourceSelectDebug(sa->resources, verbosity, indent + 2, fp);
    KheDaySelectDebug(sa->days, verbosity, indent + 2, fp);
    KheBoolDebug("priqueue", sa->priqueue, verbosity, indent, fp);
    KheBoolDebug("extra_selection", sa->extra_selection, verbosity, indent, fp);
    KheBoolDebug("expand_by_shifts", sa->expand_by_shifts, verbosity,
      indent, fp);
    KheBoolDebug("shift_pairs", sa->shift_pairs, verbosity,
      indent, fp);
    KheBoolDebug("correlated_exprs", sa->correlated_exprs, verbosity,
      indent, fp);
    KheIntDebug("daily_expand_limit", sa->daily_expand_limit, verbosity,
      indent, fp);
    KheIntDebug("daily_prune_trigger", sa->daily_prune_trigger, verbosity,
      indent, fp);
    KheIntDebug("resource_expand_limit", sa->resource_expand_limit, verbosity,
      indent, fp);
    KheIntDebug("dom_approx", sa->dom_approx, verbosity,
      indent, fp);
    fprintf(fp, "%*smain_dom_kind %s\n", indent + 2, "",
      KheDomKindName(sa->main_dom_kind));
    if( sa->cache )
      fprintf(fp, "%*scache_dom_kind %s\n", indent + 2, "",
	KheDomKindName(sa->cache_dom_kind));
    fprintf(stderr, "%*s]\n", indent, "");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_OPTIMAL_RESULT"                                           */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_OPTIMAL_RESULT KheOptimalResultMake(KHE_RESOURCE_SET rs,             */
/*    KHE_INTERVAL_SET is, KHE_COST cost, HA_ARENA a)                        */
/*                                                                           */
/*  Make an optimal result object with these attributes.                     */
/*                                                                           */
/*****************************************************************************/

static KHE_OPTIMAL_RESULT KheOptimalResultMake(KHE_RESOURCE_SET rs,
  KHE_INTERVAL_SET is, KHE_COST cost, HA_ARENA a)
{
  KHE_OPTIMAL_RESULT res;
  HaMake(res, a);
  res->resource_set = rs;
  res->day_set = is;
  res->cost = cost;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_OPTIMAL_RESULT_SET"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_OPTIMAL_RESULT_SET KheOptimalResultSetMake(HA_ARENA a)               */
/*                                                                           */
/*  Make a new, empty optimal result set.                                    */
/*                                                                           */
/*****************************************************************************/

static KHE_OPTIMAL_RESULT_SET KheOptimalResultSetMake(HA_ARENA a)
{
  KHE_OPTIMAL_RESULT_SET res;
  HaMake(res, a);
  HaArrayInit(res->results, a);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheOptimalResultSetMeldResult(KHE_OPTIMAL_RESULT_SET ors,           */
/*    KHE_RESOURCE_SET rs, KHE_INTERVAL_SET is, KHE_COST cost, HA_ARENA a)   */
/*                                                                           */
/*  Meld a new result into ors.                                              */
/*                                                                           */
/*****************************************************************************/

static void KheOptimalResultSetMeldResult(KHE_OPTIMAL_RESULT_SET ors,
  KHE_RESOURCE_SET rs, KHE_INTERVAL_SET is, KHE_COST cost, HA_ARENA a)
{
  KHE_OPTIMAL_RESULT or;  int i;

  /* search for an existing result for this rs and is, and check costs */
  HaArrayForEach(ors->results, or, i)
    if( KheResourceSetEqual(rs, or->resource_set) &&
	KheIntervalSetEqual(is, or->day_set) )
    {
      HnAssert(or->cost == cost,
	"KheOptimalResultSetMeldResult: inconsistent costs");
      return;
    }

  /* no record yet, so add a new one */
  or = KheOptimalResultMake(rs, is, cost, a);
  HaArrayAddLast(ors->results, or);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DIM_POS_OPT"                                              */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DIM_POS_OPT KheDimPosOptParse(char **str, KHE_RESOURCE_TYPE rt,      */
/*    KHE_SOLN soln, KHE_FRAME days_frame, HA_ARENA a)                       */
/*                                                                           */
/*  Parse and return a dim pos option.                                       */
/*                                                                           */
/*****************************************************************************/

static KHE_DIM_POS_OPT KheDimPosOptParse(char **str, KHE_RESOURCE_TYPE rt,
  KHE_SOLN soln, KHE_FRAME days_frame, HA_ARENA a)
{
  KHE_DIM_POS_OPT res;
  HaMake(res, a);
  res->tag = KheCharParse(str);
  switch( res->tag )
  {
    /* resource select options */
    case 'R':

      res->u.resource_select_val = KheResourceSelectParse(str, rt, soln,
	false, a);
      break;

    /* day select options */
    case 'D':

      res->u.day_select_val = KheDaySelectParse(str, days_frame, false, a);
      break;

    /* boolean options */
    case 'Q':
    case 'E':
    case 'S':
    case 'P':
    case 'C':

      res->u.bool_val = KheBoolParse(str);
      break;

    /* int options */
    case 'L':
    case 'T':
    case 'M':
    case 'A':

      res->u.int_val = KheIntParse(str);
      break;

    /* dom_kind options */
    case 'K':
    case 'J':

      res->u.dom_kind_val = KheDomKindParse(str);
      break;

    /* erroneous Z option - it's parsed separately */
    case 'Z':

      HnAbort("'Z' not sole element of last rs_drs_test dimension");
      break;

    default:

      HnAbort("unknown item name '%c' in rs_drs_test option", res->tag);
      break;
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimPosOptAddDisplay(KHE_DIM_POS_OPT opt, HA_ARRAY_NCHAR *res)    */
/*                                                                           */
/*  Add a string representation of opt to res.                               */
/*                                                                           */
/*****************************************************************************/

static void KheDimPosOptAddDisplay(KHE_DIM_POS_OPT opt, HA_ARRAY_NCHAR *res)
{
  switch( opt->tag )
  {
    /* resource select options */
    case 'R':

      KheResourceSelectAddFirstResourceSetDisplay(
	opt->u.resource_select_val, res);
      break;

    /* day select options */
    case 'D':

      KheDaySelectAddFirstIntervalSetDisplay(opt->u.day_select_val,res);
      break;

    /* boolean options */
    case 'Q':
    case 'E':
    case 'S':
    case 'P':
    case 'C':

      HnStringAdd(res, "%c%s", opt->tag, bool_show(opt->u.bool_val));
      break;

    /* int options */
    case 'L':
    case 'T':
    case 'M':
    case 'A':

      HnStringAdd(res, "%c%d", opt->tag, opt->u.int_val);
      break;

    /* dom_kind options */
    case 'K':
    case 'J':

      HnStringAdd(res, "%c%s", opt->tag, KheDomKindName(opt->u.dom_kind_val));
      break;

    default:

      HnAbort("unknown item name '%c' in rs_drs_test option", opt->tag);
      break;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DIM_POS"                                                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DIM_POS KheDimPosParse(char **str, KHE_RESOURCE_TYPE rt,             */
/*    KHE_SOLN soln, KHE_FRAME days_frame, HA_ARENA a)                       */
/*                                                                           */
/*  Parse *str to find one dim pos.                                          */
/*                                                                           */
/*****************************************************************************/

static KHE_DIM_POS KheDimPosParse(char **str, KHE_RESOURCE_TYPE rt,
  KHE_SOLN soln, KHE_FRAME days_frame, HA_ARENA a)
{
  KHE_DIM_POS res;  KHE_DIM_POS_OPT opt;
  HaMake(res, a);
  HaArrayInit(res->options, a);
  if( **str >= 'A' && **str <= 'Z' )
  {
    do
    {
      opt = KheDimPosOptParse(str, rt, soln, days_frame, a);
      HaArrayAddLast(res->options, opt);
    }
    while( KheStrSkip(str, ":") );
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimPosAddDisplay(KHE_DIM_POS dim_pos, HA_ARRAY_NCHAR *res)       */
/*                                                                           */
/*  Add a display of dim_pos to res.                                         */
/*                                                                           */
/*****************************************************************************/

static void KheDimPosAddDisplay(KHE_DIM_POS dim_pos, HA_ARRAY_NCHAR *res)
{
  KHE_DIM_POS_OPT opt;  int i;
  HaArrayForEach(dim_pos->options, opt, i)
  {
    if( i > 0 )
      HnStringAdd(res, ":");
    KheDimPosOptAddDisplay(opt, res);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDimPosDisplay(KHE_DIM_POS dim_pos, HA_ARENA a)                  */
/*                                                                           */
/*  Return a name for dim_pos based on its elements.                         */
/*                                                                           */
/*****************************************************************************/

static char *KheDimPosDisplay(KHE_DIM_POS dim_pos, HA_ARENA a)
{
  HA_ARRAY_NCHAR res;
  HnStringBegin(res, a);
  KheDimPosAddDisplay(dim_pos, &res);
  return HnStringEnd(res);
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_DIM"                                                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_DIM KheDimParse(char **str, KHE_RESOURCE_TYPE rt,                    */
/*    KHE_SOLN soln, KHE_FRAME days_frame, HA_ARENA a)                       */
/*                                                                           */
/*  Parse str and return the dim.                                            */
/*                                                                           */
/*****************************************************************************/

static KHE_DIM KheDimParse(char **str, KHE_RESOURCE_TYPE rt,
  KHE_SOLN soln, KHE_FRAME days_frame, HA_ARENA a)
{
  KHE_DIM res;  KHE_DIM_POS dim_pos;
  HaMake(res, a);
  HaArrayInit(res->dim_poses, a);
  if( KheStrSkip(str, "Z") )
  {
    HnAssert(KheParseHasEnded(str), "KheDimParse: Z not at end");
    res->days_dim = true;
  }
  else
  {
    res->days_dim = false;
    do
    {
      dim_pos = KheDimPosParse(str, rt, soln, days_frame, a);
      HaArrayAddLast(res->dim_poses, dim_pos);
    } while( KheStrSkip(str, ";") );
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDimAddDisplay(KHE_DIM dim, HA_ARRAY_NCHAR *res)                  */
/*                                                                           */
/*  Add a display of dim to res.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheDimAddDisplay(KHE_DIM dim, HA_ARRAY_NCHAR *res)
{
  KHE_DIM_POS dim_pos;  int i;
  HaArrayForEach(dim->dim_poses, dim_pos, i)
  {
    if( i > 0 )
      HnStringAdd(res, "; ");
    KheDimPosAddDisplay(dim_pos, res);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheDimDisplay(KHE_DIM dim, HA_ARENA a)                             */
/*                                                                           */
/*  Return a name for dim based on its elements.                             */
/*                                                                           */
/*****************************************************************************/

static char *KheDimDisplay(KHE_DIM dim, HA_ARENA a)
{
  HA_ARRAY_NCHAR res;
  if( dim->days_dim )
    return "Days";
  else
  {
    HnStringBegin(res, a);
    KheDimAddDisplay(dim, &res);
    return HnStringEnd(res);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "KHE_RUNNER"                                                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_RUNNER KheRunnerBegin(KHE_DYNAMIC_RESOURCE_SOLVER drs,               */
/*    KHE_SOLN soln, KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)              */
/*                                                                           */
/*  Make and return a new KHE_RUNNER object based on these attributes.       */
/*                                                                           */
/*****************************************************************************/

static KHE_RUNNER KheRunnerBegin(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_SOLN soln, KHE_RESOURCE_TYPE rt, KHE_OPTIONS options)
{
  KHE_RUNNER res;  HA_ARENA a;
  a = KheSolnArenaBegin(soln);
  HaMake(res, a);
  res->arena = a;
  res->soln = soln;
  res->resource_type = rt;
  res->options = options;
  res->days_frame = KheOptionsFrame(options, "gs_common_frame", soln);
  KheRandomGeneratorInit(&res->rgen, KheSolnDiversifier(soln));
  res->solve_args = KheSolveArgumentsGet(soln, rt, res->days_frame, options, a);
  res->solver = drs;
  HnAssert(res->solver != NULL, "KheRunnerMake: unable to make DRS solver");
  HaArrayInit(res->test_dims, a);
  res->optimal_results = KheOptimalResultSetMake(a);
  if( DEBUG3 )
    KheSolveArgumentsDebug(res->solve_args, 1, 0, stderr);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunnerEnd(KHE_RUNNER runner)                                     */
/*                                                                           */
/*  Finish up with runner.                                                   */
/*                                                                           */
/*****************************************************************************/

static void KheRunnerEnd(KHE_RUNNER runner)
{
  KheDynamicResourceSolverDelete(runner->solver);
  KheSolnArenaEnd(runner->soln, runner->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunnerParseTest(char **str, KHE_RUNNER runner)                   */
/*                                                                           */
/*  Parse one test (a sequence of two or more dimensions) and add it to      */
/*  runner.                                                                  */
/*                                                                           */
/*****************************************************************************/

static void KheRunnerParseTest(char **str, KHE_RUNNER runner)
{
  KHE_DIM dim;
  do
  {
    dim = KheDimParse(str, runner->resource_type, runner->soln,
      runner->days_frame, runner->arena);
    HaArrayAddLast(runner->test_dims, dim);
  } while( KheStrSkip(str, "|") );
  HnAssert(KheParseHasEnded(str),
    "stray characters \"%s\" at end of rs_drs_test", *str);
  HnAssert(HaArrayCount(runner->test_dims) >= 2,
    "too few | characters in rs_drs_test");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddResources(KHE_DYNAMIC_RESOURCE_SOLVER drs,                 */
/*    KHE_RESOURCE_SET rset)                                                 */
/*                                                                           */
/*  Add rset to drs.                                                         */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddResources(KHE_DYNAMIC_RESOURCE_SOLVER drs,
  KHE_RESOURCE_SET rset)
{
  int i;
  for( i = 0;  i < KheResourceSetResourceCount(rset);  i++ )
    KheDynamicResourceSolverAddResource(drs, KheResourceSetResource(rset, i));
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDrsAddDays(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_INTERVAL_SET is) */
/*                                                                           */
/*  Add is to drs.                                                           */
/*                                                                           */
/*****************************************************************************/

static void KheDrsAddDays(KHE_DYNAMIC_RESOURCE_SOLVER drs, KHE_INTERVAL_SET is)
{
  int i;  KHE_INTERVAL in;
  HaArrayForEach(is->intervals, in, i)
    KheDynamicResourceSolverAddDayRange(drs, in.first, in.last);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunnerSolve(KHE_RUNNER runner, int solve_limit, int fail_limit)  */
/*                                                                           */
/*  Perform a VLSN solve using runner, with up to solve_limit solves, and    */
/*  up to fail_limit consecutive solves that produce no improvement.         */
/*                                                                           */
/*****************************************************************************/

static void KheRunnerSolve(KHE_RUNNER runner, int solve_limit, int fail_limit)
{
  int solve_count, fail_count, i, j;  KHE_RESOURCE_SET rset;
  KHE_INTERVAL_SET iset;  KHE_SOLVE_ARGUMENTS sa;  bool success;
RESTART:
  solve_count = 0;
  fail_count = 0;
  sa = runner->solve_args;
  KheResourceSelectForEach(sa->resources, &runner->rgen, rset, i)
  {
    KheDaySelectForEach(sa->days, &runner->rgen, iset, j)
    {
      KheDrsAddResources(runner->solver, rset);
      KheDrsAddDays(runner->solver, iset);
      success = KheDynamicResourceSolverSolve(runner->solver, sa->priqueue,
	sa->extra_selection, sa->expand_by_shifts, sa->shift_pairs,
	sa->correlated_exprs, sa->daily_expand_limit,
	sa->daily_prune_trigger, sa->resource_expand_limit, sa->dom_approx,
	sa->main_dom_kind, sa->cache, sa->cache_dom_kind);
      if( success && sa->resources->kind == KHE_RESOURCE_SELECT_CLUSTER )
      {
	/* restart */
	KheResourceSelectSetCluster(sa->resources, runner->soln, runner->arena);
	goto RESTART;
      }
      solve_count++;
      fail_count = (success ? 0 : fail_count + 1);
      if( solve_count >= solve_limit || fail_count >= fail_limit ||
	  KheOptionsTimeLimitReached(runner->options) )
	return;
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunnerDoOneTest(KHE_RUNNER runner, KHE_SOLVE_ARGUMENTS sa)       */
/*                                                                           */
/*  Carry out one test using solve arguments sa (not runner->solve_args).    */
/*                                                                           */
/*****************************************************************************/

static void KheRunnerDoOneTest(KHE_RUNNER runner, KHE_SOLVE_ARGUMENTS sa)
{
  KHE_RESOURCE_SET rset;  KHE_INTERVAL_SET iset;  KHE_COST cost;
  rset = KheResourceSelectFirstResourceSet(sa->resources, &runner->rgen);
  KheDrsAddResources(runner->solver, rset);
  iset = KheDaySelectFirstIntervalSet(sa->days, &runner->rgen);
  KheDrsAddDays(runner->solver, iset);
  KheDynamicResourceSolverTest(runner->solver, sa->priqueue,
    sa->extra_selection, sa->expand_by_shifts, sa->shift_pairs,
    sa->correlated_exprs, sa->daily_expand_limit, sa->daily_prune_trigger,
    sa->resource_expand_limit, sa->dom_approx, sa->main_dom_kind,
    sa->cache, sa->cache_dom_kind, &cost);
  if( sa->daily_expand_limit == 0 && sa->daily_prune_trigger == 0 &&
      sa->resource_expand_limit == 0 && sa->dom_approx == 0 )
    KheOptimalResultSetMeldResult(runner->optimal_results, rset, iset,
      cost, runner->arena);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheFileAndGraphBegin(KHE_RUNNER runner, char *file_stem,            */
/*    char *key_label, char *below_caption, char *left_caption,              */
/*    KHE_FILE *file, KHE_GRAPH *graph, HA_ARENA a)                          */
/*                                                                           */
/*  Set *file to a new output file containing a new *graph with these        */
/*  attributes.                                                              */
/*                                                                           */
/*****************************************************************************/

static void KheFileAndGraphBegin(KHE_RUNNER runner, char *file_stem,
  char *key_label, char *below_caption, char *left_caption,
  KHE_FILE *file, KHE_GRAPH *graph, HA_ARENA a)
{
  char *file_name;
  if( strlen(file_stem) == 0 )
    file_name = HnStringMake(a, "res_%s", left_caption);
  else
    file_name = HnStringMake(a, "%s_%s", file_stem, left_caption);
  *file = KheFileBegin(file_name, KHE_FILE_LOUT_STANDALONE);
  *graph = KheGraphBegin(*file);
  KheGraphSetBelowCaption(*graph, below_caption);
  KheGraphSetLeftCaptionAndGap(*graph, left_caption, "1.0c");
  KheGraphSetKeyLabel(*graph, key_label);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheFileAndGraphEnd(KHE_FILE file, KHE_GRAPH graph)                  */
/*                                                                           */
/*  End the use of file and graph.                                           */
/*                                                                           */
/*****************************************************************************/

static void KheFileAndGraphEnd(KHE_FILE file, KHE_GRAPH graph)
{
  KheGraphEnd(graph);
  KheFileEnd(file);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunnerDoTest(KHE_RUNNER runner, int dim_index,                   */
/*    KHE_SOLVE_ARGUMENTS sa, char *graph_name, char *key_label)             */
/*                                                                           */
/*  Carry out the actual work of KheRunnerTest.                              */
/*                                                                           */
/*****************************************************************************/

static void KheRunnerDoTest(KHE_RUNNER runner, int dim_index,
  KHE_SOLVE_ARGUMENTS sa, char *graph_name, char *key_label)
{
  KHE_DIM dim, last_dim;  KHE_SOLVE_ARGUMENTS sub_sa, sub_sub_sa;
  KHE_DIM_POS dim_pos, last_dim_pos;  HA_ARRAY_NCHAR anc;
  char *sub_graph_name, *sub_graph_key_label, *data_set_name, *below_caption;
  KHE_DATASET_POINTS_TYPE points_type;
  KHE_DATASET made_kd, size_kd, time_kd;
  KHE_FILE made_file, size_file, time_file;
  KHE_GRAPH made_graph, size_graph, time_graph;
  int i, j, k, count, solns_made, table_size;
  int total_solns_made, total_table_size;  float running_time;
  if( dim_index == HaArrayCount(runner->test_dims) - 2 )
  {
    /* execute one set of graphs using sa */
    dim = HaArray(runner->test_dims, dim_index);
    last_dim = HaArray(runner->test_dims, dim_index + 1);

    /* start off the graphs */
    below_caption = KheDimDisplay(last_dim, runner->arena);
    KheFileAndGraphBegin(runner, graph_name, key_label, below_caption,
      "Made", &made_file, &made_graph, runner->arena);
    KheFileAndGraphBegin(runner, graph_name, key_label, below_caption,
      "Size", &size_file, &size_graph, runner->arena);
    KheFileAndGraphBegin(runner, graph_name, key_label, below_caption,
      "Secs", &time_file, &time_graph, runner->arena);

    /* one data set for each position along the second-last dimension */
    HaArrayForEach(dim->dim_poses, dim_pos, i)
    {
      points_type = (KHE_DATASET_POINTS_TYPE) (i % 10 + 1);
      sub_sa = KheSolveArgumentsCopyAndUpdate(sa, dim_pos, runner->arena);
      data_set_name = KheDimPosDisplay(dim_pos, runner->arena);
      made_kd = KheDataSetAdd(made_graph, points_type,
	KHE_DATASET_PAIRS_DASHED, data_set_name);
      size_kd = KheDataSetAdd(size_graph, points_type,
	KHE_DATASET_PAIRS_DASHED, data_set_name);
      time_kd = KheDataSetAdd(time_graph, points_type,
	KHE_DATASET_PAIRS_DASHED, data_set_name);

      if( last_dim->days_dim )
      {
	/* last dimension is Z - one point for each day of a single test */
	KheRunnerDoOneTest(runner, sub_sa);
	count = KheDynamicResourceSolverSolveStatsCount(runner->solver);
	for( j = 0;  j < count;  j++ )
	{
          KheDynamicResourceSolverSolveStats(runner->solver, j, &solns_made,
	    &table_size, &running_time);
	  if( count <= 28 || j % 2 == 0 || j == count - 1 )
	  {
	    KhePointAdd(made_kd, (float) j, solns_made);
	    KhePointAdd(size_kd, (float) j, table_size);
	    KhePointAdd(time_kd, (float) j, running_time);
	  }
	}
      }
      else
      {
	/* last dimension is not Z - one test for each position */
	HaArrayForEach(last_dim->dim_poses, last_dim_pos, j)
	{
	  sub_sub_sa = KheSolveArgumentsCopyAndUpdate(sub_sa, last_dim_pos,
	    runner->arena);
	  KheRunnerDoOneTest(runner, sub_sub_sa);
	  count = KheDynamicResourceSolverSolveStatsCount(runner->solver);
	  total_solns_made = 0;
	  total_table_size = 0;
	  running_time = 0;
	  for( k = 0;  k < count;  k++ )
	  {
	    KheDynamicResourceSolverSolveStats(runner->solver, k,
	      &solns_made, &table_size, &running_time);
            total_solns_made += solns_made;
            total_table_size += table_size;
	  }
	  KhePointAdd(made_kd, (float) j, total_solns_made);
	  KhePointAdd(size_kd, (float) j, total_table_size);
	  KhePointAdd(time_kd, (float) j, running_time);
	}
      }
    }

    /* finish off the graphs */
    KheFileAndGraphEnd(made_file, made_graph);
    KheFileAndGraphEnd(size_file, size_graph);
    KheFileAndGraphEnd(time_file, time_graph);
  }
  else
  {
    /* execute multiple graphs */
    dim = HaArray(runner->test_dims, dim_index);
    HaArrayForEach(dim->dim_poses, dim_pos, i)
    {
      /* make a graph_name for this position */
      if( HaArrayCount(dim->dim_poses) == 1 )
	sub_graph_name = graph_name;
      else if( strlen(graph_name) > 0 )
	sub_graph_name = HnStringMake(runner->arena, "%s.%d", graph_name, i+1);
      else
	sub_graph_name = HnStringMake(runner->arena, "%d", i + 1);

      /* make a key_label for this position */
      HnStringBegin(anc, runner->arena);
      if( strlen(key_label) > 0 )
	HnStringAdd(&anc, "%s|", key_label);
      KheDimPosAddDisplay(dim_pos, &anc);
      sub_graph_key_label = HnStringEnd(anc);
      /* ***
	sub_graph_key_label = HnStringMake(runner->arena, "%s|%s",
	  key_label, KheDimPosDisplay(dim_pos, runner->arena));
      else
	sub_graph_key_label = KheDimPosDisplay(dim_pos, runner->arena);
      *** */

      /* make sub_sa for this position along dim */
      sub_sa = KheSolveArgumentsCopyAndUpdate(sa, dim_pos, runner->arena);

      /* and execute this position */
      KheRunnerDoTest(runner, dim_index + 1, sub_sa, sub_graph_name,
	sub_graph_key_label);
    }
  }
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheStrAdjust(char *str, HA_ARENA a)                                */
/*                                                                           */
/*  Return str, adjusted to remove any colons, which are not good in file    */
/*  names.                                                                   */
/*                                                                           */
/*****************************************************************************/

static char *KheStrAdjust(char *str, HA_ARENA a)
{
  HA_ARRAY_NCHAR anc;  char *p, ch;
  HnStringBegin(anc, a);
  for( p = str;  *p != '\0';  p++ )
  {
    ch = *p;
    if( ch == ':' )
      ch = '_';
    HnStringAdd(&anc, "%c", ch);
  }
  return HnStringEnd(anc);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheRunnerTest(KHE_RUNNER runner)                                    */
/*                                                                           */
/*  Perform a test using runner.                                             */
/*                                                                           */
/*****************************************************************************/

static void KheRunnerTest(KHE_RUNNER runner)
{
  char *rt_id, *file_stem;
  rt_id = KheResourceTypeId(runner->resource_type);
  file_stem = KheStrAdjust(rt_id, runner->arena);
  KheRunnerDoTest(runner, 0, runner->solve_args, file_stem, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "main functions"                                               */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  bool KheDynamicResourceVLSNSolve(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,    */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Carry out a VLSN search using calls to KheDynamicResourceSolverSolve.    */
/*                                                                           */
/*****************************************************************************/

bool KheDynamicResourceVLSNSolve(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,
  KHE_OPTIONS options)
{
  float rs_drs_time_limit;  int rs_drs_solve_limit, rs_drs_fail_limit;
  char *str; KHE_COST init_cost;  KHE_TIMER timer;  KHE_RUNNER runner;
  KHE_DYNAMIC_RESOURCE_SOLVER drs;

  /* quit immediately if the rs_drs_off option is true */
  if( KheOptionsGetBool(options, "rs_drs_off", false) )
    return false;

  if( DEBUG1 )
    fprintf(stderr, "[ KheDynamicResourceVLSNSolve(%s, %s, options)\n",
      KheInstanceId(KheSolnInstance(soln)), KheResourceTypeId(rt));

  /* get rs_drs_time_limit and set timer */
  rs_drs_time_limit = KheTimeFromString(
    KheOptionsGet(options, "rs_drs_time_limit", "-"));
  if( rs_drs_time_limit >= 0.0 )
    timer = KheOptionsAddTimer(options, "vlsn", rs_drs_time_limit);
  else
    timer = NULL;

  /* get rs_drs_solve_limit */
  str = KheOptionsGet(options, "rs_drs_solve_limit", "1000");
  if( strcmp(str, "-") == 0 )
    rs_drs_solve_limit = INT_MAX;
  else
    rs_drs_solve_limit = KheOptionsGetInt(options, "rs_drs_solve_limit", 1000);

  /* get rs_drs_fail_limit */
  str = KheOptionsGet(options, "rs_drs_fail_limit", "20");
  if( strcmp(str, "-") == 0 )
    rs_drs_fail_limit = INT_MAX;
  else
    rs_drs_fail_limit = KheOptionsGetInt(options, "rs_drs_fail_limit", 20);

  /* create a runner, carry out the solves, and delete the runner */
  init_cost = KheSolnCost(soln);
  drs = KheDynamicResourceSolverMake(soln, rt, options);
  if( drs != NULL )
  {
    runner = KheRunnerBegin(drs, soln, rt, options);
    KheRunnerSolve(runner, rs_drs_solve_limit, rs_drs_fail_limit);
    KheRunnerEnd(runner);
  }

  /* delete the timer */
  if( timer != NULL )
    KheOptionsDeleteTimer(options, timer);

  /* all done, return */
  if( KheSolnCost(soln) < init_cost )
  {
    if( DEBUG1 )
      fprintf(stderr,
	"] KheDynamicResourceVLSNSolve returning true (%.5f < %.5f)\n",
	KheCostShow(KheSolnCost(soln)), KheCostShow(init_cost));
    return true;
  }
  else
  {
    if( DEBUG1 )
     fprintf(stderr, "] KheDynamicResourceVLSNSolve returning false%s\n",
	drs == NULL ? " (no drs)" : "");
    return false;
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheDynamicResourceVLSNTest(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,     */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Test dynamic resource solving and produce graphs of the results, as      */
/*  documented in the User's Guide.                                          */
/*                                                                           */
/*****************************************************************************/

void KheDynamicResourceVLSNTest(KHE_SOLN soln, KHE_RESOURCE_TYPE rt,
  KHE_OPTIONS options)
{
  KHE_RUNNER runner;  char *rs_drs_test;
  KHE_DYNAMIC_RESOURCE_SOLVER drs;

  /* get rs_drs_test option and quit immediately if it is "none" */
  rs_drs_test = KheOptionsGet(options, "rs_drs_test", "none");
  if( strcmp(rs_drs_test, "none") == 0 )
    return;

  /* debug */
  if( DEBUG2 )
  {
    fprintf(stderr, "[ KheDynamicResourceVLSNTest(%s, %s, options)\n",
      KheInstanceId(KheSolnInstance(soln)), KheResourceTypeId(rt));
    fprintf(stderr, "  %s\n", rs_drs_test);
  }

  /* make a runner, parse a test, run that test, and end */
  drs = KheDynamicResourceSolverMake(soln, rt, options);
  if( drs != NULL )
  {
    runner = KheRunnerBegin(drs, soln, rt, options);
    KheRunnerParseTest(&rs_drs_test, runner);
    KheRunnerTest(runner);
    KheRunnerEnd(runner);
  }

  if( DEBUG2 )
    fprintf(stderr, "] KheDynamicResourceVLSNTest\n");
}
